from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
import requests
import urlparse
import time

from wsf.session import session, messages

from PyQt4 import QtCore

class WsfProxyRequestHandler(BaseHTTPRequestHandler):
    def __init__(self, *args, **kwargs):
        BaseHTTPRequestHandler.__init__(self, *args, **kwargs)

    def __do_proxy(self):
        ''' Method that actually works as the proxy core, sends the client request to
            the remote server and sends back to the client the server response '''
        t = time.time()
        # gather the last mid available, instantiate one WsfMessage and store it
        if self.server._threaded:
            self.server._mutex.lock()
            mid = self.server.session.mid
            self.server.session.mid += 1
            m = messages.WsfMessage(mid)
            self.server.session.store(m)
            self.server._mutex.unlock()
        else:
            mid = self.server.session.mid
            self.server.session.mid += 1
            m = messages.WsfMessage(mid)
            self.server.session.store(m)

        # craft own http message
        m.req = self.__craft_request()
        # update the elapsed time (ignore time spent on crafting)
        m.elapsed = time.time() - t

        # Alter request if necessary and retrieve the modified one
        if self.server.intercept:
            self.server.session.store(m)
            self.alter_request(m.mid)
            m = self.server.session.retrieve(m.mid)

        t = time.time()
        # send to remote server client request
        if m.modified:
            m.res = self.__send_to_remote_server(m.m_req)
        else:
            m.res = self.__send_to_remote_server(m.req)

        # finish calculating the elapsed time
        m.elapsed += (time.time() - t)

        self.server.session.store(m)
        # send the server response to client
        if isinstance(m.res, messages.WsfResponse):
            # evaluate if intercept is enabled
            if self.server.intercept:
                if m.res['supported']:
                    self.alter_response(m.mid)
                    m = self.server.session.retrieve(m.mid)
                else:
                    m.m_res = m.res
                self.__send_to_client(m.m_res)
            else:
                self.__send_to_client(m.res)
        else:
            self.__request_error(-1, m.res)

        # save in the DB the data
        self.server.session.store(m)
        self.server.session.store_pres(m)
        # emit a signal to the GUI
        self.server.emit(QtCore.SIGNAL("newMessage"), m)

    # Handle all methods with __do_proxy
    do_HEAD = __do_proxy
    do_GET = __do_proxy
    do_POST = __do_proxy
    do_PUT = __do_proxy
    do_DELETE = __do_proxy

    def log_message(self, *args):
        ''' Overwrite this method so it doesn't print anything in screen '''
        return

    def __read_postdata(self):
        if 'content-length' in self.headers.keys():
            l = int(self.headers['content-length'])
            return self.rfile.read(l)

    def __craft_request(self):
        ''' Crafts a WsfRequest object from the information received '''
        u = urlparse.urlparse(self.path)

        postdata = ""
        if self.command == "POST":
            postdata = self.__read_postdata()

        req = messages.WsfRequest(method=self.command, path=u.path,
            protocol=self.protocol_version, scheme=u.scheme, netloc=u.netloc,
            url=self.path)
        req.add_query(u.query)
        req.add_headers(self.headers.headers)
        if postdata:
            req.add_data(postdata)

        return req

    def alter_request(self, mid):
        ''' Intercepts the traffic and sends the appropriate signals to the
        interface so the user can alter the request '''
        # send signal to interface
        self.server.emit(QtCore.SIGNAL("mIntercepted"), mid, "req")
        # wait on interface for 'Forward' or 'Stop' intercept
        if self.server._threaded and self.server.intercept:
            self.server._mutex.lock()
            self.server._waitC.wait(self.server._mutex)
            self.server._mutex.unlock()

    def alter_response(self, mid):
        ''' Intercepts the traffic and sends the appropriate signals to the
        interface so the user can alter the response '''
        self.server.emit(QtCore.SIGNAL("mIntercepted"), mid, "res")
        # wait on interface for 'Forward' or 'Stop' intercept
        if self.server._threaded and self.server.intercept:
            self.server._mutex.lock()
            self.server._waitC.wait(self.server._mutex)
            self.server._mutex.unlock()

    def __send_to_remote_server(self, req):
        ''' Forward the request the proxy received to the remote server '''
        headers_no_host = req['headers'].copy()
        headers_no_host.pop('host', None)
        # FIXME WHAT IF THIS FAILS??????
        res = messages.WsfResponse()
        try:
            # Make the request and craft the WsfResponse
            r = requests.request(req['method'], req['url'],
            headers=req['headers'], params=req.query_as_txt(),
            data=req.data_as_txt(), allow_redirects=False, stream=True)

            res['protocol'] = req['protocol']
            res['code'] = r.raw.status
            res['reason'] = r.raw.reason
            res['headers'] = r.headers
            res['encoding'] = r.encoding
            res['body'] = r.content
            enc = r.encoding

            # setting the size of the response (some responses won't have a
            # content-length header)
            res['length'] = len(res['body'])
            if 'content-length' in res['headers']:
                res['headers']['content-length'] = res['length']

            # evaluate if it is printable content
            if r.headers['content-type'].split('/')[0] != "text":
                res['supported'] = False
            # not needed since content won't be gziped/deflated or similar
            res['headers'].pop('content-encoding', None)

        except Exception, e:
            #FIXME if exception is cause it wont return an empty response
            # message
            res = e

        return res

    def __send_to_client(self, res):
        ''' Forward the response the proxy received from the server to the
            client requesting it '''
        self.send_response(int(res['code']))
        for h in res['headers'].keys():
            if h == 'transfer-encoding' \
                and res['headers'][h] == 'chunked':
                    # FIXME check if it was modified and update this field
                    self.send_header('Content-Length', len(res['body']))
            else:
                self.send_header(h, res['headers'][h])
        self.end_headers()

        if res['supported']:
            self.wfile.write(res['body'])
        elif res['encoding']:
            self.wfile.write(res['body'].decode(res['encoding']))
        else:
            self.wfile.write(res['body'])

    def __request_error(self, code, msg):
        ''' Returns a simple html to the browser indicating that there was an
            error performing request '''
        self.send_response(code)
        self.send_header('Connection', 'close')
        self.send_header('Content-type', 'text/html')
        self.end_headers()
        err = "WSF ERROR[%s]: Error while processing your request:<br/>%s" % (code, msg)
        self.wfile.write(err) #FIXME better message
        self.wfile.close()


class WsfProxyServer(HTTPServer):
    def __init__(self, *args, **kwargs):
        '''
            Starts an HTTPProxyServer if the following parameters are received
            ip: IP address to listen
            port: port number to listen
            ph: proxy handler to use
            dbfile: filepath to storing database
        '''
        self._ip = kwargs.pop('ip', 'localhost')
        self._port = kwargs.pop('port', 8080)
        self._ph = kwargs.pop('ph', None)
        self._dbfile = kwargs.pop('dbfile', '/var/tmp/temp.wsf')

        # flow control bools
        self._stop = False
        self._threaded = False
        self.intercept = False

        # session
        self.session = session.WsfSession(self._dbfile)

        if self._ip == '':
            self._ip = 'localhost'
        addr = (self._ip, self._port)
        HTTPServer.__init__(self, addr, self._ph, *args, **kwargs)

    def get_ip():
        ''' Return ip that the proxy is bound to '''
        return self._ip

    def get_port():
        ''' Return port that the proxy is bound to '''
        return self._port

    def toggle_intercept(self):
        ''' Toggles the intercept mode and prepares the proxy for it'''
        self.intercept = not self.intercept

