import simplejson

LINE_END = '\n'
CRLF = '\r\n'
UNWANTED_HEADERS = ['if-match', 'if-modified-since', 'if-none-match',
    'if-unmodified-since']

class WsfReqRes(dict):
    def _params_as_dict(self, params):
        ''' Given a parameters strings returns them as a dictionary '''
        d = {}
        if params != "":
            params = params.split("&")
            for param in params:
                if "=" in param:
                    param = param.split("=")
                    d[param[0].lower()] = param[1].strip()

        return d

    def _dict_as_params(self, params_dict):
        ''' Given a dictionary of parameters returns them as an string '''
        params = ""
        for k, v in params_dict.iteritems():
            param = "%s=%s&" % (k, v) # will have an extra trailing &
            params += param

        return params.rstrip("&")

    def _headers_as_dict(self, headers):
        ''' Given a list of headers returns them as a dictionary if these are
        not present in the list of UNWANTED_HEADERS '''
        d = {}
        if headers:
            for h in headers:
                h2 = h.split(':')
                if h2 and h2[0].lower() not in UNWANTED_HEADERS:
                    d[h2[0].lower()] = h2[1].strip()
        return d

    def add_headers(self, headers):
        self['headers'] = self._headers_as_dict(headers)

    def as_json(self):
        ''' Prints response in JSON (indented) '''
        return simplejson.dumps(self, indent=4)

class WsfRequest(WsfReqRes):
    def __init__(self, method="", path="", protocol="", scheme="http",
    netloc="", url=""):
        ''' Initialises the WSF HTTP Request object '''
        dict.__init__(self)
        self.setdefault('method', method)
        self.setdefault('path', path)
        self.setdefault('scheme', scheme)
        self.setdefault('netloc', netloc)
        self.setdefault('url', url)
        self.setdefault('query', {})
        self.setdefault('protocol', protocol)
        self.setdefault('headers', {})
        self.setdefault('data', {})

    def add_query(self, params):
        self['query'] = self._params_as_dict(params)

    def add_data(self, params):
        self['data'] = self._params_as_dict(params)

    def get_sn(self):
        ''' Returns the concatenation of the scheme and the netloc '''
        return "%s://%s" % (self['scheme'], self['netloc'])

    def query_as_txt(self):
        return self._dict_as_params(self['query'])

    def data_as_txt(self):
        return self._dict_as_params(self['data'])

    def as_txt(self):
        ''' Returns an string containing a valid request '''
        # if it's empty return an empty string
        if not self:
            return ""

        req = ""
        # build the first two line of the request
        query =  self._dict_as_params(self['query'])
        pathquery = self['path']
        if query:
            pathquery = "%s?%s" % (self['path'], query)
        req += "%s %s %s%s" % (self['method'], pathquery,\
                                  self['protocol'], CRLF)

        req += "Host: %s%s" % (self['headers']['host'], CRLF)

        # include the headers
        for k, v in self['headers'].iteritems():
            if k.lower() != 'host':
                req += "%s: %s%s" % (k, v, CRLF)

        # trailing line and body parameters in case those exist
        req += "%s" % CRLF
        if self['data']:
            req += "%s" % self._dict_as_params(self['data'])

        return req

    def from_txt(self, req_str):
        ''' Given a request in string format it populates a WsfRequest object '''
        # sanitise the input and convert everything to LINE_END for easy parsing
        req_str = req_str.replace(CRLF, LINE_END)
        i = 0
        is_body = False
        num_line_end = 0
        for l in req_str.split(LINE_END):
            if l == '':
                l = LINE_END

            # parse first line
            if i == 0:
                l = l.split(' ')
                self['method'] = l[0]
                self['path'] = l[1]
                url_args = l[1].split("?")
                if len(url_args) > 1 : # url has params
                    self['path'] = url_args[0] # rebuild the path
                    self.add_query(url_args[1]) # add the query params
                self['protocol'] = l[2]
            # rest of the body
            else:
                if l == LINE_END:
                    num_line_end += 1
                    if num_line_end == 1:
                        is_body = True
                    elif num_line_end > 1:
                        is_body = False

                if l != LINE_END:
                    if is_body: # the body itself
                        self.add_data(l)
                    elif not is_body: # it's a normal header line
                        l = l.split(":")
                        if l[0] not in UNWANTED_HEADERS:
                            self['headers'][l[0].lower()] = l[1].strip() # build headers dict
            i += 1
        self['scheme'] = 'http' # not supporting anything else at the moment
        self['netloc'] = self['headers']['host']
        self['url'] = "%s://%s%s?%s" % (self['scheme'], self['netloc'], \
                        self['path'], self.query_as_txt())


    def as_json(self):
        ''' Prints request in JSON (indented) '''
        return simplejson.dumps(self, indent=4)

class WsfResponse(WsfReqRes):
    def __init__(self):
        ''' Initialises the WSF HTTP Response object '''
        dict.__init__(self)
        self.setdefault('code', -1)
        self.setdefault('protocol', "")
        self.setdefault('reason', "")
        self.setdefault('headers', {})
        self.setdefault('body', "")
        self.setdefault('length', 0)
        self.setdefault('supported', True)
        self.setdefault('encoding', None)

    def as_txt(self):
        res = self._headers_as_txt()
        res += self._body_as_txt()
        return res

    def _headers_as_txt(self):
        res = "%s %s %s%s" % (self['protocol'], self['code'], self['reason'], CRLF)
        for k, v in self['headers'].iteritems():
                res += "%s: %s%s" % (k, v, CRLF)
        res += "%s" % CRLF
        return res

    def _body_as_txt(self):
        if self['supported'] and self['body']:
            return self['body']
        return "Content not printable"

    def from_txt(self, req_str):
        ''' Given a response in string format it populates a WsfRequest object '''
        # sanitise the input and convert everything to LINE_END for easy parsing
        req_str = req_str.replace(CRLF, LINE_END)
        i = 0
        is_body = False
        num_line_end = 0
        splitted_req = req_str.split(LINE_END)
        for l in splitted_req:
            if l == '':
                l = LINE_END

            # parse first line
            if i == 0:
                l = l.split(' ')
                self['protocol'] = l[0]
                self['code'] = int(l[1])
                self['reason'] = " ".join(l[2:])
            # rest of the body
            else:
                if l == LINE_END:
                    num_line_end += 1
                    if num_line_end == 1:
                        is_body = True

                if l != LINE_END:
                    if is_body: # the body itself
                        self['body'] = "\n".join(splitted_req[i:])
                        break
                    elif not is_body: # it's a normal header line
                        l = l.split(":")
                        if l[0] not in UNWANTED_HEADERS:
                            self['headers'][l[0].lower()] = l[1].strip() # build headers dict
            i += 1

        # update supported field
        self['supported'] = True
        if 'content-type' in self['headers']:
            if self['headers']['content-type'].split('/')[0] != "text":
                self['supported'] = False

        self['length'] = len(self['body'])
        if 'content-length' in self['headers']:
            self['headers']['content-length'] = self['length']
        # update length field

class WsfMessage(object):
    def __init__(self, mid, req=None, res=None):
        self.mid = mid # WsfMessage ID
        self.modified = False
        self.req = req or WsfRequest()
        self.res = res or WsfResponse()
        # m_req and m_res will hold the potentially tampered requests/responses
        self.m_req = None
        self.m_res = None
        # holds the value of the time elapsed
        self.elapsed = 0

    def as_gui(self):
        # FIXME evaluate that req/res are not empty
        if not self.modified:
            has_params = False or self.req['query'] != {} \
                    or self.req['data'] != {}
            m_gui = (self.mid, self.req['headers']['host'], \
                self.req['method'], self.req['url'], str(has_params), \
                str(self.res['code']), str(self.res['length']), \
                str(self.modified), str(round(self.elapsed*1000, 1)))
        else:
            has_params = False or self.m_req['query'] != {} \
                    or self.m_req['data'] != {}
            m_gui = (self.mid, self.m_req['headers']['host'], \
                self.m_req['method'], self.m_req['url'], str(has_params), \
                str(self.m_res['code']), str(self.m_res['length']), \
                str(self.modified), str(round(self.elapsed*1000, 1)))

        return m_gui

