from PyQt4 import QtGui, QtCore, uic, Qt

import os
import shutil

from wsf.http.proxy import WsfProxyServer, WsfProxyRequestHandler
from wsf.session.messages import WsfRequest, WsfResponse
from wsf.session.session import WsfSession

(Ui_WSF, QMainWindow) = uic.loadUiType('./wsf/ui/wsf.ui')

class WsfQtProxyServer(QtCore.QThread, WsfProxyServer):
    def __init__(self, *args, **kwargs):
        WsfProxyServer.__init__(self, *args, **kwargs)
        QtCore.QThread.__init__(self)
        self._threaded = True
        self._mutex = Qt.QMutex()
        self._waitC = Qt.QWaitCondition()

    def run(self):
        while not self._stop:
            self.handle_request()

class WsfGui(QtGui.QMainWindow):
    def __init__(self, parent=None, *args, **kwargs):
        ''' Finishes the tweaking of the UI: connects signals with slots, edits
        few properties, ...'''
        QtGui.QMainWindow.__init__(self, parent)
        self.ui = Ui_WSF()
        self.ui.setupUi(self)

        # Setup Proxy
        self._ip = kwargs.pop('ip', 'localhost')
        self._port = kwargs.pop('port', 8080)
        self._dbfile = kwargs.pop('dbfile', '/var/tmp/temp.wsf')
        self.proxy = self.__init_proxy()
        # message that is being intercepted and modified
        self._intercepted_m = None
        self._intercepted_typ = None

        # Several adjustments before initialising the GUI
        self.__init_gui()

        # bind signals with slots
        self.__connections()
        # start proxy
        self.proxy.start()

    def __init_proxy(self):
        # Initialize proxy handler
        ph = WsfProxyRequestHandler
        # Initialize Qt Proxy and start serving
        proxy = WsfQtProxyServer(ip=self._ip, port=self._port, ph=ph,\
            dbfile=self._dbfile)

        return proxy

    def __set_status(self, dbfile):
        status = "Proxy listening at: %s:%s\t|\tSession file: %s" % (self._ip, \
        self._port, dbfile)
        self.ui.statusBar.showMessage(status)

    def __connections(self):
        # Own defined SIGNALS
        # Signal for new messages in the session
        self.connect(self.proxy, QtCore.SIGNAL("newMessage"), self.sl_new_message)

        # Signal for request/response intercepted
        self.connect(self.proxy, QtCore.SIGNAL("mIntercepted"), self.sl_m_intercept)

        ##########
        # Qt default SIGNALS
        # Requests widget: row is selected
        self.ui.requestsWidget.itemSelectionChanged.connect(self.sl_handle_rw)

        # Intercept: start/stop button is clicked
        self.ui.btInterceptStart.toggled.connect(self.sl_start_stop_intercept)
        # Intercept: forward button is clicked
        self.ui.btForward.clicked.connect(self.sl_m_forward)

        ##########
        # Menu bindings
        # File - Load
        self.ui.actionLoad_session.triggered.connect(self.sl_menu_load)
        # File - Save
        self.ui.actionSave_session.triggered.connect(self.sl_menu_save)

        # About
        self.ui.actionAbout.triggered.connect(self.sl_menu_about)

    def closeEvent(self, event):
        if self.proxy.session.changes():
            msg = "The current session appears to contain data. Do you " \
            "want to save it first?"
            r = QtGui.QMessageBox.question(self, "Message", msg, \
                    QtGui.QMessageBox.Yes, QtGui.QMessageBox.No)

            if r == QtGui.QMessageBox.Yes:
                self.sl_menu_save()

        # remove the default file
        try:
            os.remove(self._dbfile)
        except os.error:
            # silently fail to remove the file
            pass

    # #############################
    # Slots
    @QtCore.pyqtSlot()
    def sl_new_message(self, m):
        ''' Slot requests widget: new message arrives'''
        self.ins_message([m.as_gui()])

    @QtCore.pyqtSlot()
    def sl_handle_rw(self):
        ''' Slot requests widget: row is selected '''
        # clear the text areas
        hReq = self.ui.hReqText
        hRes = self.ui.hResText
        hReqMod = self.ui.hReqModText
        hResMod = self.ui.hResModText
        hReq.clear()
        hRes.clear()
        hReqMod.clear()
        hResMod.clear()

        # get the selected row number
        r = self.ui.requestsWidget.selectedIndexes()[0].row()
        # read the first value for that row and the first colum (MID)
        mid = int(self.ui.requestsWidget.item(r, 0).text())
        m = self.proxy.session.retrieve(mid)
        hReq.insertPlainText(QtCore.QString(m.req.as_txt()))
        hRes.insertPlainText(QtCore.QString(m.res.as_txt()))
        if m.modified:
            hReqMod.insertPlainText(QtCore.QString(m.m_req.as_txt()))
            hResMod.insertPlainText(QtCore.QString(m.m_res.as_txt()))

    @QtCore.pyqtSlot()
    def sl_start_stop_intercept(self):
        ''' Intercept: start/stop button is clicked '''
        # update GUI
        if self.ui.btInterceptStart.isChecked():
            self.ui.btInterceptStart.setText("Stop")
        else:
            self.ui.btInterceptStart.setText("Start")
        self.proxy.toggle_intercept()
        # if a request/response was intercepted release it
        self.proxy._waitC.wakeAll()
        # clear the interface
        self.ui.interceptText.clear()

    @QtCore.pyqtSlot()
    def sl_m_intercept(self, mid, typ):
        ''' Either a request or a response has been received in the intercept
        mode, update the interface '''
        iText = self.ui.interceptText
        iText.clear()
        # retrieve message
        m = self.proxy.session.retrieve(mid)
        # update control fields
        self._intercepted_m = m
        self._intercepted_typ = typ

        if typ == "req":
            s = QtCore.QString(m.req.as_txt())
        else:
            s = QtCore.QString(m.res.as_txt())
        iText.insertPlainText(s)

    @QtCore.pyqtSlot()
    def sl_m_forward(self):
        ''' Forward button is pressed in the interface '''
        iText = self.ui.interceptText

        # if there was a message intercepted update it with the changes
        if self._intercepted_m:
            # mark it as modified
            self._intercepted_m.modified = True
            if self._intercepted_typ == "req":
                f_str = str(iText.toPlainText())
                self._intercepted_m.m_req = WsfRequest()
                self._intercepted_m.m_req.from_txt(f_str)
            else:
                enc = self._intercepted_m.res['encoding']
                if enc:
                    f_str = unicode(iText.toPlainText()).encode(enc)
                else:
                    f_str = str(iText.toPlainText())
                self._intercepted_m.m_res = WsfResponse()
                self._intercepted_m.m_res.from_txt(f_str)
            self.proxy.session.store(self._intercepted_m)
        self._intercepted_m = None
        # stop the proxy from waiting
        self.proxy._waitC.wakeAll()
        # clear the interface
        iText.clear()

    @QtCore.pyqtSlot()
    def sl_menu_load(self):
        ''' Load session menu is selected '''
        sess = Qt.QFileDialog.getOpenFileName(self, "Open session file", \
            os.getenv('HOME'), "WSF session files (*.wsf)")
        # a file name was provided
        sess = str(sess)
        if sess:
            # ask for saving changes if there is anything no -> remove,
            # yes -> save dialog
            if self.proxy.session.changes():
                msg = "The current session appears to contain data. Do you " \
                "want to save it first?"
                r = QtGui.QMessageBox.question(self, "Message", msg, \
                        QtGui.QMessageBox.Yes, QtGui.QMessageBox.No)

                if r == QtGui.QMessageBox.Yes:
                    self.sl_menu_save()

            curr = self.proxy.session._dbfile
            try:
                self.proxy.session = WsfSession(sess)
                curr = sess
            except Exception:
                Qt.QMessage.information(self, "Error", "Couldn't load session")
                self.proxy.sesion = WsfSession(curr)
            # Show items in the session
            self.__set_status(curr)
            p = self.proxy.session.retrieve_pres()
            self.ui.requestsWidget.setRowCount(0)
            self.ins_message(p)

    @QtCore.pyqtSlot()
    def sl_menu_save(self):
        ''' Save session menu is selected '''
        # show only the dialog if there is anything to save
        if self.proxy.session.changes():
            sess = Qt.QFileDialog.getSaveFileName(self, "Save session", \
            os.getenv('HOME'), "WSF session files (*.wsf)")
            sess = str(sess)
            if not sess.endswith(".wsf"):
                sess += ".wsf"

            curr = self.proxy.session._dbfile
            try:
                shutil.copy(curr, sess)
                self.proxy.session = WsfSession(sess)
            except Exception:
                Qt.QMessage.information(self, "Error", "Couldn't save session")

    @QtCore.pyqtSlot()
    def sl_menu_about(self):
        Qt.QMessageBox.about(self, "About", "WSF v0.1a (prototype)\nAdrian " \
        "Hermoso Metaute\nUniversitat de Barcelona")

    # #############################
    # Different UI tweaks
    def __init_gui(self):
        ''' Tweaks several widgets properties before the GUI is presented to the
        user '''
        rW = self.ui.requestsWidget # Alias for the requests widget
        # rw items non-editable
        rW.setEditTriggers(QtGui.QAbstractItemView.NoEditTriggers)

        # load items in the session if there were any in the temp file
        p = self.proxy.session.retrieve_pres()
        self.ins_message(p)

        # updates status bar
        self.__set_status(self._dbfile)


    # #############################
    # IO
    def ins_message(self, requests):
        ''' Given an array of specifically formated requests updates the UI with
        it '''
        rW = self.ui.requestsWidget # Alias for the requests widget

        # increase the rowcount
        rows = rW.rowCount()
        rW.setRowCount(rows + len(requests))

        rW.setSortingEnabled(False)
        for i, r in enumerate(requests):
            for j, col in enumerate(r):
                item = QtGui.QTableWidgetItem(col)
                # make first column a list of integers so it can be sorted
                if j == 0:
                    item.setData(QtCore.Qt.DisplayRole, col)
                rW.setItem(rows+i, j, item)
        rW.setSortingEnabled(True)
        #TODO get selected header column or default to mid col#0
        rW.sortItems(0)

    # #############################
